
:- import append/3, member/2, ith/3, length/2 from basics.
:- import subsumes/2 from subsumes.

% The corresponding routines for CONDITIONAL FOLDING

cond_fold(Folded, P/N, NewClause) :-
	!, find_folder_clauses(P/N, Folder),
	cond_fold(Folded, Folder, NewClause).

cond_fold(Folded, Folder, NewClause) :-
	retractall(before_gen(_,_,_)),
	( gen_check(Folded)     % To generalize or not to generalize
		-> cond_fold_gen(Folded, Folder, NewClause)
		;  cond_fold_normal(Folded, Folder, NewClause)
        ).

cond_fold_normal(Folded, Folder, NewClause) :-
 	( cond_foldable(Folded, Folder, C, Biject) 
		-> apply_fold(C, Biject, NewClause)
                ;  fail
        ).

cond_fold_gen(Folded, Folder, NewClause) :-
	( cond_foldable(Folded, Folder, C, Biject) 
		-> apply_fold(C, Biject, NewClause)
                ;  restore_non_gen(Folded), fail
        ).


%--------------------------------------------------------------
% cond_foldable(Folded, Folder, C, Biject)
% ---  performs foladability check modulo replacement of conditions
% ---  these conditions are proved correct inside cond_foldable/4


cond_foldable(Folded, Folder, C, Biject) :-
	permute(Folded, Fd), 
	cond_inst_all(Fd, Folder, Clist, Blist, NewFd), 
	Biject = (NewFd, Folder), 
	variant_all(Clist), variant_all(Blist), 
	Clist = [C|_], Blist = [B|_],
	soundnessCheck(B, Folder), !.


% allows only conjunctive conditions

cond_inst_all([Cd|FdRest],[Cr|FrRest], [C|CRest],[B|BRest], NewFolded) :-
	( inst_one(Cd, Cr, C, B) 
                 -> cond_inst_all(FdRest,FrRest, CRest,BRest, Folded1),
		    NewFolded = [Cd|Folded1]
		 ;  inst_all(FdRest,FrRest,CRest,BRest),
                    cond_inst_one(Cd, Cr, C, B, Cd1),
		    NewFolded = [Cd1|FdRest]
        ).


cond_inst_one(Cd, Cr, New, Hr, NewId) :-
	get_clause(Cd, (Hd :- Bd), (Annot_d1, Annot_d2)),
	get_clause0(Cr, (Hr :- Br), (_Annot_r1, Annot_r2)),
	find_internal_vars((Hr :- Br), IVars),
	fold_cond_select(Bd,Br,IVars,Bd_nf,Conj1,Conj2,Fd_part),
	Clause_d = clauseI(Cd, (Hd :- Bd), (Annot_d1, Annot_d2)),
	get_vars(Conj1, Vars1), get_vars(Conj2, Vars2),
	var_setminus(Vars1,IVars,HV1), var_setminus(Vars2,IVars,HV2),
	goal_replace(Conj1/HV1, Conj2/HV2, Clause_d, NewId),
        subsumes(Br, Fd_part), check_internal_inst(IVars, [Hd|Bd_nf]),
	append(Bd_nf, [Hr], NewBody),
	New = (Hd :- NewBody),
	clauseI(NewId, _, (AdNew, _)),  % \delta has been added by now
	compute_vector(Bd_nf, CV),
	plus_annot(AdNew, CV, Ad1), 
	greater_annot(Ad1, Annot_r2).

%----------------------------------------------------------------------

fold_cond_select(Bd, Br, IVars, Bd_nf, Conj1, Conj2, Fd_part) :-
	match_folder(Br, Bd, Bd_nf, Bd_f), 
	cond_subsumes(Br, Bd_f, IVars, Conj1, Conj2),
	remove_all(Conj1,Bd_f,Bd_f1), append(Bd_f1,Conj2,Fd_part).


% cond_subsumes(Br, Bd, IVars, Conj1, Conj2)
% ---  succeeds if : 
% ---      after replacing Conj1 by Conj2 in folded cl. body Bd,
% ---      Bd subsumes Br. 

cond_subsumes(Br, Bd, IVars, Conj1, Conj2) :-
	find_insts(Br, Bd, Insts), shared_var_triples(Br, 1, SVT),
	wrong_subsuming_atoms(Insts, Insts, IVars, 1, AList1),
	sharing_conflict(SVT, Insts, IVars, AList2),
	append(AList1, AList2, AList3), remove_dupl(AList3, AList),
	length(AList, M), fold_cond_max(N), M < N,
	replaced_conj(Bd, AList, Conj1, Sort_AList),
	replacing_conj(Br, Bd, Sort_AList, Conj2).

%-----------------------------------------------------------------------

shared_var_triples([], _, []).

shared_var_triples([A|Rest], N, SVT) :-
	M is N + 1, 
	shared_var_triples(A, Rest, N, M, SVT1),
	shared_var_triples(Rest, M, SVT2),
	append(SVT1, SVT2, SVT).

:- index shared_var_triples/5-2.

shared_var_triples(_A, [], _, _, []).

shared_var_triples(A, [B|Rest], NA, NB, SVT) :-
	shared_vars_all(A, B, NA, NB, SVT1), 
	NB1 is NB + 1,
	shared_var_triples(A, Rest, NA, NB1, SVT2),
	append(SVT1, SVT2, SVT).

shared_vars_all(A, B, NA, NB, SVT) :-
	find_shared_vars(A, B, VList), 
	var_remove_dupl(VList, VList1),
	construct_triples(VList1, NA, NB, SVT).

construct_triples([], _, _, []).
construct_triples([V|Rest], NA, NB, [share(V, NA, NB) | TRest]) :-
	construct_triples(Rest, NA, NB, TRest).

%-----------------------------------------------------------------

wrong_subsuming_atoms([], _, _, _, []).

wrong_subsuming_atoms([Inst|IRest], Insts, IVars, NA, AList) :-
	NA1 is NA + 1,
	wrong_subsuming_atoms(IRest, Insts, IVars, NA1, AList1),
	( (Inst == null ; wrong_inst_local_vars(IVars, Inst, Insts))
		 -> AList = [NA|AList1]
		 ;  AList = AList1
        ).


find_insts([], [], []).
find_insts([A|Rest], [B|Rest1], [Inst|IRest]) :-
	find_inst(A, B, Inst),
	find_insts(Rest, Rest1, IRest).

find_inst(A, B, Inst) :-
	( subsumes(A, B, Inst) -> Inst = Inst
		               ;  Inst = null
        ).


%------------------------------------------------------------------

:- dynamic sharing_coeff/2.

sharing_conflict(SVT, Insts, IVars, AList) :-
	detect_sharing_conflict(SVT, Insts, IVars, SVT1),
	compute_sharing_coeff(SVT1),
	group_sharing_conflict(SVT1, SVT2),
	resolve_sharing_conflict(SVT2, AList),
	retractall(sharing_coeff(_, _)).
                               

detect_sharing_conflict([], _Insts, _IVars, []).

detect_sharing_conflict([share(X,I,J)|Rest], Insts, IVars, Conflict) :-
	( var_mem(X, IVars) -> C = []
		            ;  ( retrieve_inst(X, I, Insts, T1),
			         retrieve_inst(X, J, Insts, T2),
			         ( T1==T2 -> C = []
				          ;  C = [share(X, (I,T1), (J,T2))]
                                 )
                               )
        ), detect_sharing_conflict(Rest, Insts, IVars, Conflict1),
	append(C, Conflict1, Conflict).


retrieve_inst(X, I, Insts, T1) :-
	ith(I, Insts, Insts_I), member((Y = T1), Insts_I), X == Y.


compute_sharing_coeff([]).

compute_sharing_coeff([H|Tail]) :-
	H = share(_, (I, _), (J,_)),
	update_sharing_coeff(I),
	update_sharing_coeff(J),
	compute_sharing_coeff(Tail).	

update_sharing_coeff(I) :-
	( sharing_coeff(I, N) -> retractall(sharing_coeff(I, _)),
		         	 M is N + 1,
			         assert(sharing_coeff(I, M))
			      ;  assert(sharing_coeff(I, 1))		    
        ).

group_sharing_conflict([], []).

group_sharing_conflict([share(X, I, J)|Rest], [share(X, List1)|SVT1]) :-
	locate_similar_conflict(X, Rest, Rest1, List),
	var_remove_dupl([I,J|List], List1),
	group_sharing_conflict(Rest1, SVT1).

locate_similar_conflict(X, List, Rest, Conflicts) :-
	locate_similar_conflict(X, List, [], Rest, Conflicts).

locate_similar_conflict(_X, [], Rest, Rest, []).

locate_similar_conflict(X, [H|Tail], SoFar, Rest, Conflicts) :-
	H = share(Y, I, J), 
	( X == Y -> Temp = [I, J], SoFar1 = SoFar
                 ;  Temp = [], SoFar1 = [H|SoFar]
        ), locate_similar_conflict(X, Tail, SoFar1, Rest, C1),
	append(Temp, C1, Conflicts).
	

resolve_sharing_conflict([], []).

resolve_sharing_conflict([share(_X, List)|Rest], AList) :-
	sharing_sort(List, List1),
	select_sharing_inst(List1, Removed_Atoms),
	resolve_sharing_conflict(Rest, AList1),
	append(AList1, Removed_Atoms, AList).
	

select_sharing_inst(List, Removed) :-
	member((_I, T), List),  % List is "sorted"; so inst. 
				% selected in preferred order.
	sharing_based_removal(List, T, Removed).
	
sharing_based_removal([], _, []).

sharing_based_removal([(I,T1)|Rest], T, Removed) :-
	(T == T1 -> Temp = [] 
		 ;  Temp = [I] 
	), sharing_based_removal(Rest, T, Rem1),
	append(Temp, Rem1, Removed).


sharing_sort([], []).

sharing_sort([(I, T)|Rest], List) :-
	sharing_sort(Rest, List1),
	sharing_insert((I, T), List1, List).

sharing_insert((I, T), [], [(I, T)]).

sharing_insert((I, T1), [(J, T2)|Rest], List) :-
	sharing_coeff(I, N_i), sharing_coeff(J, N_j), 
	( N_i > N_j -> sharing_insert((I,T1), Rest, List1),
		       List = [(J, T2)|List1]
		    ;  (N_i == N_j -> depth(T1, D1), depth(T2, D2),
				      ( D1 > D2 -> List = [(J,T2),(I,T1)|Rest]
		                                ;  List = [(I,T1),(J,T2)|Rest]
                                      )
			           ;  List = [(I,T1),(J,T2)|Rest]
                       )
        ).


%---------------------------------------------------------------------------

replaced_conj(Bd, AList, Conj, AList1) :-
	int_sort(AList, AList1), retrieve_elements(Bd, AList1, Conj, _).
	
replacing_conj(Br, Bd, AList, Conj) :-
	retrieve_elements(Bd, AList, _, Bd_rm),
	retrieve_elements(Br, AList, Repl, Br_rm),
	subsumes(Br_rm, Bd_rm, Inst),
	copy_term(Repl, Conj, VMap),
	map_term(VMap, Inst, Map_Inst),
	apply_subst(Map_Inst).

%---------------------------------------------------------------------------

