
:- import str_cat/3 from string.
:- import conset/2, conget/2 from gensym.
:- import length/2, member/2 from basics.

format(PgmFile) :-
	load_dyn(PgmFile), conset(label, 0), write_clauses.

write_clauses :- 
	setof(P1/K1, pgm_pred(P1/K1), PredList),
	member(P/K, PredList), 
	length(List, K), Hd =.. [P|List],
	\+ predicate_property(Hd, built_in),
	write_pred_clauses(Hd),
	findall(C, clause0(C, (Hd :- _), _), CList),
	assert(defn_pred(P/K, CList)), fail.

write_clauses :-
	clause(annot(X), B), assert((X :- B)), 
        retract((annot(X) :- B)), fail.

write_clauses :-
	recursive_pred(P/N), length(L, N), Hd =.. [P|L],
	assert(inductive(Hd)), fail.

write_clauses :- 
	toplevel(P/N), assert(proof_oblig(P/N)), fail.

write_clauses.


write_pred_clauses(Hd) :-
	clause(Hd, Bd), conget(label, N), commaConvert(Bd, Bd1),
	assert(clauseI(clause(N), (Hd :- Bd1), (1,1))), 
	assert(clause0(clause(N), (Hd :- Bd1), (1,1))), 
	M is N + 1, conset(label, M), fail.
		
write_pred_clauses(Hd) :- my_retractall(Hd).

my_retractall(Hd) :- clause(Hd, Bd), retract((Hd :- Bd)), fail.
my_retractall(_).


%--------------------------------------------------------------------

recursive_pred(P/N) :-
	pgm_pred(P/N), pred_tc(P/N, P/N).

pgm_pred(Q/M) :- toplevel(P/N), tc(P/N, Q/M).

% compute transitive closure of predicate dependencies.

edge(P/N, Q/M) :-
	length(L, N), Head =.. [P|L],
	\+ predicate_property(Head, built_in),
	clause(Head, Body), commaConvert(Body, Body1),
	member(X, Body1), functor(X, Q, M),
	\+ predicate_property(Head, built_in).

:- table tc/2.

tc(X, X).

tc(X, Y) :- tc(X, Z), edge(Z, Y).


commaConvert(L1, L2) :-
	is_list(L1) -> L2 = L1 ; commaConvert1(L1, L2).

commaConvert1(CL, VL) :- 
	( CL = (G,Rest) -> commaConvert1(Rest,VL1), VL = [G|VL1]
                        ;  ( CL = true -> VL = []
                                       ;  VL = [CL]
                           )
        ).