
% cond_foldable(Folded, Folder, C_fd, C, Biject)
% --- Succeeds if after replacing some atom Atom1 by another atom
%     Atom2 in any one of the folded clauses.
%     it is possible to fold folded clause-set Folded by folder 
%     clause-set Folder (using folding bijection Biject) so as
%     to produce clause C.

cond_foldable(Folded, Folder, C, Biject) :-
	permute(Folded, Fd), 
	Biject = (Fd, Folder),
	cond_inst_all(Fd, Folder, Clist, Blist), 
	variant_all(Clist), variant_all(Blist), 
	Clist = [C|_], Blist = [B|_],
	soundnessCheck(B, Folder), !.
	

%--- Now the corresponding checks for applicability of conditional folding

cond_inst_all([], [], _, _) :- fail.

cond_inst_all([Cd|FdRest],[Cr|FrRest],[C|CRest],[B|BRest]) :-
	( inst_one(Cd, Cr, C, B) 
                 -> cond_inst_all(FdRest,FrRest,CRest,BRest)
		 ;  cond_inst_one(Cd, Cr, C, B),
		    inst_all(FdRest,FrRest,CRest,BRest)
        ).


% cond_inst_one(Cd, Cr, New, Hr)
% --- Cd,Cr are folded/folder clauses. Succeeds if :
%     using the equivalence Atom1 \equiv Atom2 (which is also proved)
%     clause Cd can be folded by clause Cr to produce clause 
%     New (containing folder atom Hr).

cond_inst_one(Cd, Cr, New, Hr) :-
	get_clause(Cd, (Hd :- Bd)),
	get_clause0(Cr, (Hr :- Br), (_, Ar)),
	find_internal_vars((Hr :- Br), IVars),
	fold_cond_select(Bd, Br, Bd_nf, Conj1, Conj2),
	goal_replace(Conj1, Conj2, Cd, Cd1),
	check_internal_inst(IVars, [Hd|Bd_nf]),
	append(Bd_nf, [Hr], NewBody),
	New = (Hd :- NewBody),
	clauseI(Cd1, _, (AdNew, _)),   % \delta has been added by now
	compute_vector(Bd_nf, CV),
	plus_annot(AdNew, CV, Ad1), 
	greater_annot(Ad1, Ar).

%------------------------------------------------------------
% the groundwork for conditional folding
% Check for subsumption of two clause bodies modulo difference
% in conjunctions of a certain pre-specified maximum lentgh. 
% Return the instantiation of the variables
% as well as the one atom in which the two bodies are different.

cond_subsumes(Bd_r, Bd_d, (Conj_d, Conj_r)) :-
	\+ subsumes(Bd_r, Bd_d),  
	remove(Conj_r, Bd_r, Bd_rm_r), % remove an atom from folder clause
	remove(Conj_d, Bd_d, Bd_rm_d), % remove an atom from folded clause
	get_vars(Conj_r, Vars_atm), 
	get_vars(Bd_rm_r, Vars_rest),
	var_subseteq(Vars_atm, Vars_rest),
        subsumes(Bd_rm_r, Bd_rm_d).

%--------------------------------------------------------------------
% fold_cond_select is used to select conditions to enable folding

fold_cond_select(Bd, Br, Bd_nf, Conj1, Conj2) :- 
	match_folder(Br, Bd, Bd_nf, Bd_f), 
	cond_subsumes(Br, Bd_f, (Conj1, Conj2)).


match_folder([], Bd, Bd, []).

match_folder([A|Rest], Bd, Bd_nf, [X|Bd_f1]) :-
	pred_match_member(A, Bd, X, Bd1),
	match_folder(Rest, Bd1, Bd_nf, Bd_f1).
	
pred_match_member(A, L, X, Rest) :-
	pred_match_member(A, L, [], X, Rest).

pred_match_member(A, [B|Tail], Prev, X, Rest) :-
	( functor(A, P, N), functor(B, P, N) 
		-> X = B, append(Prev, Tail, Rest)
		; append(Prev, [B], Prev1), 
		  pred_match_member(A, Tail, Prev1, X, Rest)
        ).

