:- import member/2, append/3, length/2 from basics.
:- dynamic consumer/1.
:- dynamic always_unfold/1.

select_strategy(G, _Num, CId) :-
	unfold_flag(suspend_expand), !, 
	( always_chk(G) ;
	  ( (functor(G, F, N), consumer(F/N))
             -> \+ open_atom(G)
	     ;  depth_check(G),
		ancestor_check(G, CId)
                %sharedVar_check(G, CId)
          )
        ).
	
% Note that for consumer preds, assuming you start unfolding an 
% open atom, you are always guaranteed to consume the existing binding
% and get back to an open atom. So it is safe to unfold any occurrence
% of a consumer pred. except an open atom.

select_strategy(G, Num, CId) :-
	unfold_flag(suspend_freeze), !,
	( always_chk(G) ;
	  ( depth_check(G), consumer_chk(G),
	    \+ unfold_prohibited(CId, Num), % subsumes the "ancestor_check"
	    sharedVar_check(G, CId)
          )
        ).

% less restrictive selection strategy allowing more aggressive unfolding
% select_strategy(G, CId) :-
%	depth_check(G), ancestor_check(G, CId), \+ unfold_prohibited(G).


fold(ClauseList, [NewClause|NFdList]) :-
	find_folder_clauses(FolderList), 
	find_folded_clauses(ClauseList, FolderList, FdList, NFdList),
	fold(FdList, FolderList, NewClause).

cond_fold(ClauseList, [NewClause|NFdList]) :-
	find_folder_clauses(FolderList), 
	find_folded_clauses(ClauseList, FolderList, FdList, NFdList),
	cond_fold(FdList, FolderList, NewClause).

%-------------------------------------------------------------------
% Find clause sets which are likely to be folded clauses for 
% these folder clauses. Right now a coarse selection strategy
% used for both conditional and unconditional folding.
% permuting the folded clauses for folding
% i.e. computing the folding bijection

find_folded_clauses(CList, Folder, FdList, NFdList) :-
	permute_clauses(Folder, CList, FdList, NFdList, Nf_atoms_List),
	prop_match_all(Nf_atoms_List).  % only check for propositional
                                        % matching at this stage.
				        % Foldability check done later.

permute_clauses([], NFdList, [], NFdList, []).

permute_clauses([D|Rest], L, [C|Fd1], NFdList, [Nf|NfRest]) :-
	match_clause(D, L, C, L1, Nf),
	permute_clauses(Rest, L1, Fd1, NFdList, NfRest).

match_clause(D, List, Clause, RemClauses, Nf) :-
	clause0(D, (_ :- Br), _),
	member(Clause, List),  
	clauseI(Clause, (_ :- Bd), _),
	remove_builtins(Br, Br1),
	remove_builtins(Bd, Bd1),
	prop_match(Br1, Bd1, Nf),
	remove(Clause, List, RemClauses).


remove_builtins([], []).
remove_builtins([G|Rest], List) :-
	remove_builtins(Rest, List1),
	( predicate_property(G, built_in) 
		-> List = List1
		;  List = [G|List1]
        ).

prop_match([], Rem, Rem).

prop_match([A|Rest], List, Rem) :-
	prop_match(A, List, [], List1),
	prop_match(Rest, List1, Rem).
			

prop_match(A, [B|Rest], SoFar, Rem) :-
	( functor(A, P, N), functor(B, P, N)
		-> append(SoFar, Rest, Rem)
		;  append(SoFar, [B], SoFar1),
		   prop_match(A, Rest, SoFar1, Rem)
        ).


%-----------------------------------------------------------

